#include "data_loader.h"
#include <QNetworkReply>
#include <QNetworkRequest>
#include <QDebug>
#include <QFile>
#include <QDomDocument>
#include <QNetworkProxy>

const QString str_traffic_info("traffic-info");
const QString str_bridges("bridges");
const QString str_bridge("bridge");
const QString str_highways("highways");
const QString str_highway("highway");

const QString str_city_load("city-load");
const QString str_current("current");

const QString str_name("name");
const QString str_level("level");
const QString str_direction("direction");
const QString str_title("title");

CDataLoader::CDataLoader(QObject *parent) : QObject(parent),
	m_manager( new QNetworkAccessManager( this )),
	m_timer( new QTimer( this )),
	m_makeSubst( false ),
	m_uiProxyPort( 0 )
{
	bool bRes = connect( this, &CDataLoader::dataUrlChanged, this, &CDataLoader::updateData );
	Q_ASSERT(bRes);
	bRes = connect(m_timer, &QTimer::timeout, this, &CDataLoader::updateData);
	Q_ASSERT(bRes);
	bRes = connect(m_manager, &QNetworkAccessManager::finished, this, &CDataLoader::dataLoaded);
	Q_ASSERT(bRes);
	bRes = connect(this, &CDataLoader::makeSubstitutionChanged, this, &CDataLoader::loadSubstition);
	Q_ASSERT(bRes);

	bRes = connect(this, &CDataLoader::proxyChanged, this, &CDataLoader::changeProxy);
	Q_ASSERT(bRes);
}

CDataLoader::~CDataLoader()
{
	qDeleteAll(m_streets);
	m_streets.clear();
}


void CDataLoader::changeProxy()
{
	if (/*m_bUseProxy && */ !m_strProxyName.isEmpty() && m_uiProxyPort != 0)
	{
		Q_ASSERT(m_manager);
		QNetworkProxy proxy;
		proxy.setHostName(m_strProxyName);
		proxy.setPort(m_uiProxyPort);
		proxy.setType(QNetworkProxy::HttpProxy);
		m_manager->setProxy(proxy);
	}
}

void CDataLoader::setProxyHostName(const QString& _host)
{
	m_strProxyName = _host;
	Q_EMIT proxyChanged();
}

QString CDataLoader::proxyHostName() const
{
	return m_strProxyName;
}

void CDataLoader::setProxyHostPort(uint port)
{
	m_uiProxyPort = port;
	Q_EMIT proxyChanged();
}

uint CDataLoader::proxyHostPort() const
{
	return m_uiProxyPort;
}

void CDataLoader::setDataUrl(const QUrl&newUrl)
{
	m_url = newUrl;
	Q_EMIT dataUrlChanged();
}

QUrl CDataLoader::dataUrl() const
{
	return m_url;
}

void CDataLoader::setUpdateTime(unsigned int updateTime)
{
	m_timer->start( updateTime );
}

unsigned int CDataLoader::updateTime() const
{
	if( m_timer && m_timer->isActive() )
		return m_timer->interval();
	return 0;
}

bool CDataLoader::makeSubstitution() const
{
	return m_makeSubst;
}

void CDataLoader::setMakeSubstitution(bool subst)
{
	m_makeSubst = subst;
	Q_EMIT makeSubstitutionChanged();
}

QUrl CDataLoader::substitutionUrl() const
{
	return m_substUrl;
}

void CDataLoader::setSubstitutionUrl(const QUrl&url)
{
	m_substUrl = url;
	Q_EMIT makeSubstitutionChanged();
}

unsigned int CDataLoader::streetsCount() const
{
	return m_streets.size();
}

unsigned int CDataLoader::directionsCount() const
{
	int count = 0;
	foreach( const Street* str, m_streets )
	{
		count += str->streetCountDirections();
	}
	return count;
}

QString CDataLoader::getStreetTitle(unsigned int streetIndex)
{
	Q_ASSERT( (int)streetIndex < m_streets.size() );
	if( !m_substitutions.isEmpty() && m_substitutions.contains( m_streets[streetIndex]->streetName() ))
		return m_substitutions.value( m_streets[streetIndex]->streetName() );
	return m_streets[streetIndex]->streetTitle();
}

unsigned int CDataLoader::getStreetCountDirections(unsigned int streetIndex)
{
	Q_ASSERT( (int)streetIndex < m_streets.size() );
	return m_streets[streetIndex]->streetCountDirections();
}

unsigned int CDataLoader::getDirectionLevel(unsigned int streetIndex, unsigned int directionIndex)
{
	Q_ASSERT( (int)streetIndex < m_streets.size() );
	Street* street = m_streets[streetIndex];
	Q_ASSERT( directionIndex < street->streetCountDirections() );

	return street->directions()[directionIndex]->loadLevel();
}

QString CDataLoader::getDirectionName(unsigned int streetIndex, unsigned int directionIndex)
{
	Q_ASSERT( (int)streetIndex < m_streets.size() );
	Street* street = m_streets[streetIndex];
	Q_ASSERT( directionIndex < street->streetCountDirections() );

	if( !m_substitutions.isEmpty() && m_substitutions.contains( street->directions()[directionIndex]->loadName() ) )
		return m_substitutions.value( street->directions()[directionIndex]->loadName() );
	return street->directions()[directionIndex]->loadDirectionTitle();
}

void CDataLoader::updateData()
{
	m_manager->get(QNetworkRequest( m_url ) );
}

void CDataLoader::dataLoaded(QNetworkReply* reply )
{
	const char* functionName = "CDataLoaded::dataLoaded";
	const char* invalidInputFileError = "Input file is invalid";
	Q_UNUSED(functionName);
	Q_UNUSED(invalidInputFileError);

	if (reply->error() != QNetworkReply::NoError)
	{
		qDebug() << reply->errorString();
		reply->deleteLater();
		return;
	}


	QDomDocument domDocument;
	bool bRes = domDocument.setContent(reply, true);
	Q_ASSERT_X( bRes, functionName, invalidInputFileError );

	const QDomNode trafficInfoNode = domDocument.firstChildElement(str_traffic_info);
	if( trafficInfoNode.isNull() )
		return;
	const QDomNode cityLoadNode = trafficInfoNode.firstChildElement(str_city_load);
	Q_ASSERT_X( !cityLoadNode.isNull(), functionName, invalidInputFileError );

	m_uiCityLoadLevel = cityLoadNode.toElement().attribute(str_current).toUInt(&bRes);
	Q_ASSERT_X( bRes, functionName, invalidInputFileError );

	const QDomNode bridgesNode = trafficInfoNode.firstChildElement(str_bridges);
	const QDomNode highwaysNode = trafficInfoNode.firstChildElement(str_highways);
	Q_ASSERT_X( !bridgesNode.isNull() || !highwaysNode.isNull(), functionName, invalidInputFileError );


	auto fillStreetsMaps = [this]( const QDomNode& streets )
	{
		for( int iStreet = 0; iStreet < streets.childNodes().size() ; iStreet++ )
		{
			const QDomNode domNodeStreet = streets.childNodes().at( iStreet );

			Street* currStreet = new Street(domNodeStreet.toElement().attribute(str_name), domNodeStreet.toElement().attribute(str_title) );

			for( int iLoad = 0; iLoad < domNodeStreet.childNodes().size(); iLoad++ )
			{
				const QDomElement load = domNodeStreet.childNodes().at(iLoad).toElement();
				if( load.isNull() )
					continue;
				StreetLoad* streetLoad = new StreetLoad(load.attribute(str_name), load.attribute(str_level).toUInt(), load.attribute(str_direction) );
				currStreet->addStreetLoad( streetLoad );
			}
			m_streets.push_back( currStreet );
		}
	};

	qDeleteAll(m_streets);
	m_streets.clear();

	if( !bridgesNode.isNull() )
		fillStreetsMaps( bridgesNode );

	if( !highwaysNode.isNull() )
		fillStreetsMaps( highwaysNode );

	reply->deleteLater();
	domDocument.clear();
	
	Q_EMIT dataChanged();

}

void CDataLoader::loadSubstition()
{
	if( !m_substitutions.isEmpty() )
		m_substitutions.clear();

	if( !m_substUrl.isValid() || !m_makeSubst )
		return;

	QFile substFile( m_substUrl.toLocalFile() );
	if( !substFile.exists() )
		return;
	if( !substFile.open( QIODevice::ReadOnly ) )
		return;

	QDomDocument domDocument;
	bool bRes = domDocument.setContent( &substFile, true );
	if( !bRes )
	{
		substFile.close();
		return;
	}

	const QDomNode subTrafficInfoNode = domDocument.firstChildElement(str_traffic_info);
	if( subTrafficInfoNode.isNull() )
	{
		substFile.close();
		return;
	}

	const QDomNode subBridgesNode = subTrafficInfoNode.firstChildElement(str_bridges);
	const QDomNode subHighwaysNode = subTrafficInfoNode.firstChildElement(str_highways);
	if( subBridgesNode.isNull() && subHighwaysNode.isNull() )
	{
		substFile.close();
		return;
	}

	auto fillStreetsMaps = [this]( const QDomNode& streets )
	{
		for( int iStreet = 0; iStreet < streets.childNodes().size() ; iStreet++ )
		{
			const QDomNode domNodeStreet = streets.childNodes().at( iStreet );

			m_substitutions.insert(domNodeStreet.toElement().attribute(str_name),domNodeStreet.toElement().attribute(str_title) );
			for( int iLoad = 0; iLoad < domNodeStreet.childNodes().size(); iLoad++ )
			{
				const QDomElement load = domNodeStreet.childNodes().at(iLoad).toElement();
				m_substitutions.insert(load.attribute(str_name),load.attribute(str_direction) );

			}
		}
	};

	if( !subBridgesNode.isNull() )
		fillStreetsMaps( subBridgesNode );

	if( !subHighwaysNode.isNull() )
		fillStreetsMaps( subHighwaysNode );

	substFile.close();
	domDocument.clear();
}

